import unittest

from open_weather_forecast.info_extractor.http_decorator import auto_tries


class AutoTriesDecoratorTest(unittest.TestCase):
    def setUp(self):
        self.url = 'http://api.openweathermap.org/data/2.5/forecast/city?q={}'.format("London,uk")

    def tearDown(self):
        pass

    def zero_retries_test(self):
        @auto_tries(ValueError, tries=0, delay=0)
        def test_func():
            raise ValueError()

        self.assertEqual(test_func(), None)

    def zero_retries_no_exception_test(self):
        @auto_tries(ValueError, tries=0, delay=0)
        def test_func():
            return 1

        self.assertEqual(test_func(), None)

    def retries_no_exception_test(self):
        @auto_tries(ValueError, tries=1, delay=0)
        def test_func():
            return 1

        self.assertEqual(test_func(), 1)

    def retries_exception_test(self):
        @auto_tries(BaseException, tries=2, delay=0)
        def test_func():
            raise BaseException

        self.assertEqual(test_func(), None)

    def retries_wrong_exception_test(self):
        @auto_tries(ValueError, tries=2, delay=0)
        def test_func():
            raise KeyError

        self.assertRaises(KeyError, test_func)





import requests
from requests import RequestException
from sqlalchemy import create_engine

from open_weather_forecast.info_extractor.http_decorator import auto_tries
from open_weather_forecast.info_extractor.get_info_abstract import GetInfoAbstract
from open_weather_forecast.conf.global_settings import get_global_settings


class GetInfo(GetInfoAbstract):

    def __init__(self):
        self.engine = None
        self.password = None
        self.username = None
        self.host = None
        self.db_name = None
        self.db_url = None

    def filter_information(self, info_retrieved, schema, result=None):
        """
        :param info_retrieved: Info retrieved from the API
        :param schema: Schema defined by the user to extract partial information
        :param result: Structure containing the filtered information from the API
        :return:
        """
        if result is None:
            result = dict()

        for key in schema:
            if isinstance(info_retrieved, dict):
                value = info_retrieved.get(key, {})
            else:
                for element, sub_info in zip(info_retrieved, schema*len(info_retrieved)):
                    result.append(type(element)())
                    self.filter_information(element, sub_info, result[-1])
                continue

            info_type = type(schema.get(key)) if isinstance(schema.get(key), (list, dict)) else schema.get(key)
            if info_type is not type(value):
                msg = "Types from information schema and the information retrieved does not match {} {}"
                raise ValueError(msg.format(type(schema.get(key)), type(value)))

            if isinstance(value, (list, dict)):
                result[key] = type(value)()
                self.filter_information(info_retrieved.get(key), schema.get(key), result.get(key))
            else:
                if isinstance(result.get(key, None), list):
                    result[key].append(value)
                else:
                    result[key] = value
        return result

    @auto_tries(RequestException, tries=4, delay=1)
    def http_retrieve(self, url=""):
        r = requests.get(url=url)
        if r.ok:
            return False, r.json()
        else:
            return True, {}

    def get_info(self, url="", information_schema=None):
        error, temperatures_history = self.http_retrieve(url)

        if not temperatures_history or error:
            return True, {}
        else:
            try:
                if information_schema:
                    temperatures_history = self.filter_information(temperatures_history, information_schema)
                return False, temperatures_history
            except (ValueError, KeyError):
                return True, {}

    def store_data(self, data):
        raise NotImplementedError()

    def get_db_connection(self):
        gs = get_global_settings()
        update = False
        for key in ["password", "username", "host", "db_name"]:
            if gs.get(key) != getattr(self, key):
                setattr(self, key, gs.get(key))
                update = True

        if update:
            self.db_url = 'postgresql+pg8000://{username}:{password}@{host}/{db_name}'.format(
                username=self.username,
                password=self.password,
                host=self.host,
                db_name=self.db_name
            )

            self.engine = create_engine(self.db_url)
            self.engine.connect()


from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Float

Base = declarative_base()


class ForecastWeatherInfo(Base):
    __tablename__ = 'forecast_weather_info'
    dt_txt = Column(String, primary_key=True, autoincrement=False)





from open_weather_forecast.info_extractor.get_info import GetInfo


class GetTemperature(GetInfo):

    def __init__(self):
        super(GetTemperature).__init__()
        self.db = None

    def store_data(self, data):
        self.get_db_connection()
        import ipdb
        ipdb.set_trace()




import os
from datetime import datetime, timedelta
import yaml

from open_weather_forecast.conf.constants import CFG_CACHE_TTL, CFG_FILE_PATH, CFG_FILE_PATH_ENV


class SettingCache():
    """ The settings cache! Borg pattern in here
    """
    __shared_state = {}

    def __init__(self, update_callback=None):
        self.__dict__ = self.__shared_state
        if update_callback is not None:
            self.update_callback = update_callback
        if not hasattr(self, 'last_update') or ((datetime.utcnow() + timedelta(seconds=self.refresh)) > self.last_update):
            self.update()

    def update(self):
        """Use as well for manual updates
        """
        self.refresh = CFG_CACHE_TTL
        self.last_update = datetime.now()
        self.cache = self.update_callback()


def get_global_settings(refresh=False):
    """Store Global config into cache to avoid to high disk access
        Cache updates every 5 secs by default
        refresh = True => Force a cache's update
    """
    sc = SettingCache()
    if refresh:
        sc.update()
    return sc.cache


def read_settings():
    """ Read global configuration yaml file
    """
    try:
        f_path = os.environ.get(CFG_FILE_PATH_ENV).strip()
        with open(f_path, 'r') as f:
            return yaml.load(f)
    except AttributeError:
        try:
            with open(CFG_FILE_PATH, 'r') as f:
                return yaml.load(f)
        except:
            msg = "Can't find settings in %s neither %s" % (CFG_FILE_PATH_ENV, CFG_FILE_PATH)
            raise Exception(msg)

SettingCache(read_settings)




import unittest

from open_weather_forecast.conf.constants import WEATHER_INFORMATION_SCHEMA
from open_weather_forecast.info_extractor.get_temperature import GetTemperature


class GetTemperatureTest(unittest.TestCase):
    def setUp(self):
        self.url = 'http://api.openweathermap.org/data/2.5/forecast/city?q={}'.format("London,uk")

    def tearDown(self):
        pass

    def basic_test(self):
        pass
        # get_temp_manager = GetTemperature()
        # info = get_temp_manager.http_retrieve(url=self.url)
        # get_temp_manager.filter_information(info, WEATHER_INFORMATION_SCHEMA)


from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Float

Base = declarative_base()


class WeatherInfo(Base):
    __tablename__ = 'weather_info'
    dt_txt = Column(String, primary_key=True, autoincrement=False)


class Temperatures(Base):
    __tablename__ = 'temperatures'
    id = Column(Integer, primary_key=True, autoincrement=True)
    max = Column(Float, nullable=False)
    min = Column(Float, nullable=False)
    average = Column(Float, nullable=False)



import os
import logging

from epic.sales_epicenter.configuration import get_global_settings
from epic.sales_epicenter.constants import LOGGING_FORMATTER


loggers = {}


def create_logger(target):
    """Creates a new logger"""
    gs = get_global_settings()
    path = gs["logging"]["path"]
    level = gs["logging"]["level"]
    filename = os.path.join(path, "app.log")

    handler = logging.FileHandler(filename=filename)
    handler.setFormatter(logging.Formatter(LOGGING_FORMATTER))
    logger = logging.getLogger(target)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    extra = {"target": target}
    logger = logging.LoggerAdapter(logger, extra)
    return logger


def logger(target='SE'):
    """Returns logger by target
    """
    try:
        return loggers[target]
    except KeyError:
        loggers[target] = create_logger(target)
        return loggers[target]



from unittest import mock, TestCase
from unittest.mock import patch
import json
from os import path

import requests

from open_weather_forecast.conf.constants import WEATHER_INFORMATION_SCHEMA
from open_weather_forecast.info_extractor.get_info import GetInfo


class FilterInformationTest(TestCase):
    def setUp(self):
        current_directory = path.abspath(path.dirname(__file__))
        with open(path.join(current_directory, "fixtures/temperature_data.json")) as f:
            self.data = json.load(f)

    def tearDown(self):
        pass

    def complex_test(self):
        results = GetInfo().filter_information(self.data, WEATHER_INFORMATION_SCHEMA)
        self.assertEqual(results,  dict(
            list=[{'dt_txt': '2015-05-22 15:00:00', 'main': {'temp_max': 293.74, 'temp': 293.74, 'temp_min': 293.437}},
                  {'dt_txt': '2015-05-22 18:00:00', 'main': {'temp_max': 293.99, 'temp': 293.99, 'temp_min': 293.744}},
                  {'dt_txt': '2015-05-22 21:00:00', 'main': {'temp_max': 290.95, 'temp': 290.95, 'temp_min': 290.77}},
                  {'dt_txt': '2015-05-23 00:00:00', 'main': {'temp_max': 288.94, 'temp': 288.94, 'temp_min': 288.817}},
                  {'dt_txt': '2015-05-23 03:00:00', 'main': {'temp_max': 287.95, 'temp': 287.95, 'temp_min': 287.885}},
                  {'dt_txt': '2015-05-23 06:00:00',
                   'main': {'temp_max': 287.805, 'temp': 287.805, 'temp_min': 287.805}},
                  {'dt_txt': '2015-05-23 09:00:00',
                   'main': {'temp_max': 291.638, 'temp': 291.638, 'temp_min': 291.638}},
                  {'dt_txt': '2015-05-23 12:00:00',
                   'main': {'temp_max': 292.527, 'temp': 292.527, 'temp_min': 292.527}},
                  {'dt_txt': '2015-05-23 15:00:00', 'main': {'temp_max': 291.49, 'temp': 291.49, 'temp_min': 291.49}},
                  {'dt_txt': '2015-05-23 18:00:00',
                   'main': {'temp_max': 290.318, 'temp': 290.318, 'temp_min': 290.318}},
                  {'dt_txt': '2015-05-23 21:00:00',
                   'main': {'temp_max': 288.242, 'temp': 288.242, 'temp_min': 288.242}},
                  {'dt_txt': '2015-05-24 00:00:00',
                   'main': {'temp_max': 285.211, 'temp': 285.211, 'temp_min': 285.211}},
                  {'dt_txt': '2015-05-24 03:00:00',
                   'main': {'temp_max': 282.244, 'temp': 282.244, 'temp_min': 282.244}},
                  {'dt_txt': '2015-05-24 06:00:00',
                   'main': {'temp_max': 284.407, 'temp': 284.407, 'temp_min': 284.407}},
                  {'dt_txt': '2015-05-24 09:00:00',
                   'main': {'temp_max': 291.455, 'temp': 291.455, 'temp_min': 291.455}},
                  {'dt_txt': '2015-05-24 12:00:00',
                   'main': {'temp_max': 294.094, 'temp': 294.094, 'temp_min': 294.094}},
                  {'dt_txt': '2015-05-24 15:00:00', 'main': {'temp_max': 295.21, 'temp': 295.21, 'temp_min': 295.21}},
                  {'dt_txt': '2015-05-24 18:00:00',
                   'main': {'temp_max': 292.971, 'temp': 292.971, 'temp_min': 292.971}},
                  {'dt_txt': '2015-05-24 21:00:00',
                   'main': {'temp_max': 290.712, 'temp': 290.712, 'temp_min': 290.712}}])
                          )

    def basic_test(self):
        schema = {
            "message": float
        }

        results = GetInfo().filter_information(self.data, schema)
        self.assertEqual(results, {"message": 0.0139})

    def dict_as_arg_test(self):
        schema = {
            "message": float
        }
        results = {}
        GetInfo().filter_information(self.data, schema, results)
        self.assertEqual(results, {"message": 0.0139})


class HttpRetrieveTest(TestCase):

    def ok_test(self):
        answer = {"Blah": True}
        mocked = mock.MagicMock()
        mocked.ok = True
        mocked.json = mock.MagicMock(return_value=answer)

        with patch.object(requests, 'get', return_value=mocked) as mock_method:
            error, res = GetInfo().http_retrieve(url="")
            self.assertEqual(res, answer)
            self.assertEqual(error, False)

    def fail_test(self):
        mocked = mock.MagicMock()
        mocked.ok = False

        with patch.object(requests, 'get', return_value=mocked) as mock_method:
            error, res = GetInfo().http_retrieve(url="")
            self.assertEqual(res, {})
            self.assertEqual(error, True)


RETRIES_LIMIT = 3

# Configuration TTL in secs
CFG_CACHE_TTL = 5

# Configuration envar name that contains settings path
CFG_FILE_PATH_ENV = "SALES_EPICENTER_CFG"
CFG_FILE_PATH = "/etc/openweather/settings.yaml"

# Log format
LOGGING_FORMATTER = "%(asctime)s[%(levelname)s][%(target)s]%(message)s"

WEATHER_INFORMATION_SCHEMA = {
    "list": [{
                 "main": {"temp": float,
                          "temp_min": float,
                          "temp_max": float
                          },
                 "dt_txt": str
             }]
}




import time
from functools import wraps


def auto_tries(exception_to_catch, tries=4, delay=3):
    """
    Decorator to auto_tries the weather info

    :param tries:
    :param delay: Expressed in seconds, the waiting time between tries
    :return:
    """

    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            for internal_tries in range(tries):
                try:
                    return f(*args, **kwargs)
                except exception_to_catch:
                    time.sleep(delay)
            return None

        return f_retry

    return deco_retry



from abc import ABCMeta, abstractmethod


class GetInfoAbstract(metaclass=ABCMeta):
    """
    Abstract class to retrieve information from a Json API
    """

    @abstractmethod
    def store_data(self, data):
        pass

    @abstractmethod
    def http_retrieve(self, url):
        pass

    @abstractmethod
    def get_info(self, url, information_schema=None):
        pass

    @abstractmethod
    def get_db_connection(self):
        pass

